/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
"use strict";

var EXPORTED_SYMBOLS = ['DictExtension'];

Components.utils.import("chrome://dict/content/pref.js");
Components.utils.import("chrome://dict/content/dictclient.js");

if ("undefined" == typeof(DictExtension)) {
    var DictExtension = {};
};

// DictExtension.Listener class, which is passed in as the "listener" parameter
// to DictExtension.Client.defineWord().
DictExtension.Listener = function(host, port, window) {
    this.host   = host;
    this.port   = port;
    this.window = window;

    // Determine which application we are running under.
    const THUNDERBIRD_ID = "{3550f703-e582-4d05-9a08-453d09bdfdc6}";
    const SEAMONKEY_ID   = "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}";

    var appInfo = Components.classes["@mozilla.org/xre/app-info;1"].getService(
        Components.interfaces.nsIXULAppInfo);

    this.isThunderbird = false;
    this.isSeaMonkey   = false;
    switch (appInfo.ID) {

    case THUNDERBIRD_ID:
        this.isThunderbird = true;
        break;

    case SEAMONKEY_ID:
        this.isSeaMonkey = true;
        break;
    }


    // Determine the settings which affect the UI
    this.resizeIsDisabled = false;
    var pref = Components.classes["@mozilla.org/preferences-service;1"].getService(
        Components.interfaces.nsIPrefBranch);
    try {
        this.resizeIsDisabled = pref.getBoolPref('dom.disable_window_move_resize');
    }
    catch (e) {
    }
}

// wraps some HTML inside HTML/BODY tag and adds a title
DictExtension.Listener.prototype =
{
    // escapes character entities in str so that it's safe for inclusion in HTML.
    escapeHtml : function(str)
    {
        var htmlEntities = {
            "&" : "&amp;",
            '"' : "&quot;",
            "'" : "&#39;",
            "<" : "&lt;",
            ">" : "&gt;"
        };

        return str.replace(/[&''""<>]/g, function(match) htmlEntities[match]);
    },

    // wraps HTML inside HTML/BODY tag and adds a title
    createHtmlDocument : function (title, body)
    {
        var htmlDocument =
            '<!DOCTYPE html>\n' +
            '<HTML>\n'    +
            '  <HEAD>\n'  +
            '      <TITLE>' + this.escapeHtml(title) + '</TITLE>\n' +
            '      <META http-equiv="Content-Type" content="text/html; charset=utf-8">\n' +
            '      <SCRIPT type="text/javascript">\n' +

            // This HTML must be self-resizing.  Doing it from the code that
            // creates the window doesn't work with data URIs.
            'function load() {\n' +
            '    window.sizeToContent();\n' +
            '    if (window.outerWidth < window.innerWidth) {\n' +

                     // The window became narrower than the content
                     // This should never happen, but it may be happening
                     // because of bug #7097--a bug in DownloadStatusBar
                     // when it has over 40 items.
                     // I cannot fix DownloadStatusBar, but I can be defensive
                     // and resize the window back to its original width.
            '        window.outerWidth = window.innerWidth;\n' +
            '    }\n' +

                 // If the page is too long to display without scrollbars, then shorten it
            '    var proposedHeight = window.innerHeight + 70;\n' +
            '    if (600 < proposedHeight) {\n' +
            '        window.innerHeight = 600;\n' +

                     // Make the window wider to account for the width of the
                     // vertical scrollbar.  If we don't do this, a
                     // horizontal scrollbar will appear.
            '        window.innerWidth += 25;\n' +
            '    }\n' +

            '}\n' +
            '\n' +
            'function crossReference(uri) {\n' +
            '    window.open(uri, \'\', \'status=no,toolbar=no,menubar=no,location=no,scrollbars=yes,resizable=yes\');\n' +
            '    return false;\n' +
            '}\n' +
            '      </SCRIPT>\n' +
            '  </HEAD>\n' +
            '  <BODY class="dict-page" onload="load();">\n' +
            body +
            '  </BODY>\n' +
            '</HTML>\n';

        return htmlDocument;
    },


    //
    // The next three functions implement the DictExtension.Listener interface
    //

    // Called when the DictExtension.Client has definitions
    onHaveDefinitions : function (word, definitions)
    {
        // Format the definitions into HTML
        var htmlBody = '';

        // count the number of sources -- used for formatting
        var totalSources = 0;
        for (var source in definitions) {
            totalSources++;
        }

        var sourceIndex = 0;
        for (source in definitions) {

            htmlBody += '<B><DIV class="dict-database">';
            htmlBody += this.escapeHtml(source);
            htmlBody += '</DIV></B>\n';

            htmlBody += '<CODE style="white-space:pre;"><SPAN class="dict-definition">';

            var i;
            for (i = 0; i < definitions[source].length; i++) {

                var deftext = definitions[source][i].text;

                // Normalize all definitions so that they end in a single newline
                deftext = deftext.replace(/\s*$/, '\n');

                // Escape HTML entities
                deftext = this.escapeHtml(deftext);

                // Words that show up between backslashes are generally
                // pronunciation guides.  See:
                // http://goodies.xfce.org/projects/applications/xfce4-dict
                deftext = deftext.replace(
                    /\\(\S*?)\\/g,
                    '<SPAN style="color:rgb(0,127,0)"><SPAN class="dict-pronunciaton">' +
                        '$1' +
                        '</SPAN></SPAN>');

                // Words that are cross-referenced by WordNet show up
                // in braces. e.g., "{word}" and "{word(P)}".
                if (this.isThunderbird) {
                    // Thunderbird doesn't support protocol providers, so we'll just
                    // remove the surrounding curly braces so that the words don't
                    // look weird.
                    deftext = deftext.replace(/{(.*?)(\(.*\))?}/g, '$1');
                }
                else {
                    // On Firefox, where the dict Protocol provider exists, we can
                    // turn these into links.
                    var uriBase = 'dict://' + this.host + ':' + this.port + '/d:';
                    deftext = deftext.replace(
                        /{(.*?)(\(.*\))?}/g,
                        function(entireMatch, word, hint) {
                            var uri = uriBase + encodeURIComponent(DictExtension.utf8ToUcs4(word));
                            return '<a ' +
                                'href="' + uri + '" ' +
                                'onclick="return crossReference(\'' + uri + '\');"' +
                                '>' + word + '</a>';
                        });
                }

                // Append an extra newline after everything but the
                // the final definition from the final source.
                // (Otherwise there'd be a strange-looking blank space
                // at the bottom of the page.)
                if (sourceIndex + 1 != totalSources ||
                    i + 1           != definitions[source].length) {
                    deftext += '\n';
                }

                htmlBody += deftext;
            }

            htmlBody += '</SPAN></CODE>\n';
            sourceIndex++;
        }

        // create an HTML document
        var htmlDocument = this.createHtmlDocument(word, htmlBody);

        // create a window for the HTML document
        var features = [
            'status=no',
            'toolbar=no',
            'menubar=no',
            'location=no',
            'scrollbars=yes',
            'resizable=yes',
        ];
        if (this.resizeIsDisabled) {
            // If the page won't be able to resize itself, then pick a
            // width/height that will be large enough to show the definition.
            features.push('width=535');
            features.push('height=500');
        }
        else {
            // If the page will resize itself.  It looks better if the window
            // starts small and resizes itself to fit the content.
            features.push('width=475');
            features.push('height=100');
        }

        if (this.isThunderbird) {
            // Add the 'chrome' flag on Thunderbird.
            //
            // In Thunderbird 17 on Fedora Core, attempting to create a window
            // without the 'chrome' flag crashes the application. This is too
            // bad, because the chrome flag disables the context menu, which
            // prevents further lookups.
            features.push('chrome');
        }

        // Create a data URI that contains the HTML document.
        var htmlDocumentAsDataUri =
            'data:text/html;charset=UTF-8,' +
            encodeURIComponent(DictExtension.utf8ToUcs4(htmlDocument));

        // open the popup window

        // Use XPCNativeWrapper to fix a potential security vulnerability of
        // calling a window.open() that was overridden by a malicious
        // entity.
        //
        // This also works around an incompatability with
        // Tabbrowser Extension's "single window mode" where the
        // window.open() that it provides does not have a document
        // property.
        var winWrapper = new XPCNativeWrapper(this.window, 'open()');
        winWrapper.open(htmlDocumentAsDataUri, '_blank', features.join(','));
    },

    // Called when the DictExtension.Client has closest matches
    onHaveClosestMatches : function (word, matches)
    {
        // There is more than one closest match
        // Display them all and let the user decide which to define
        var matchesArray = [];
        for (var match in matches) {
            matchesArray.push(match);
        }

        var query = word + '|' + matchesArray.join('|');

        // open the chrome window
        this.window.open(
            'chrome://dict/content/close-matches.xul?' + encodeURIComponent(query),
            '',
            'chrome,scrollbars,resizable');
    },

    onNoDefinitionsFound : function (word)
    {
        this.window.alert(
            'No definitions were found for "' +
            DictExtension.utf8ToUcs4(word) +
            '".');
    },

    onError : function (error)
    {
        this.window.alert(DictExtension.utf8ToUcs4(error));
    }
}


DictExtension.defineWord = function(word, window)
{
    if (word.length == 0) {
        window.alert("Please select a word to define.");
        return;
    }

    try {
        // read the host and port from the preferences
        var host = DictExtension.getDefaultServerName();
        var port = DictExtension.getDefaultServerPort();

        var listener = new DictExtension.Listener(host, port, window);
        var client   = new DictExtension.Client();
        client.defineWord(host, port, word, listener);
    }
    catch (ex) {
        window.alert(ex);
    }
}
